# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
import tarfile
from cStringIO import StringIO

from osxcollector.output_filters.analyze import AnalyzeFilter
from osxcollector.output_filters.base_filters.output_filter import _run_filter

from amira.results_uploader import FileMetaInfo
from amira.s3 import S3Handler
from amira.sqs import SqsHandler


class AMIRA():
    """Runs the automated analysis based on the new elements in an S3
    bucket:
        1. Receives the messages from the SQS queue about the new
           objects in the S3 bucket.
        2. Retrieves the objects (OSXCollector output files) from the
           bucket.
        3. Runs the Analayze Filter on the retrieved OSXCollector
           output.
        4. Uploads the analysis results.

    JIRA integration is optional. If any of the JIRA parameters
    (`jira_server`, `jira_user`, `jira_password` or `jira_project`)
    is not supplied or `None`, attaching the analysis results to a JIRA
    issue will be skipped.

    :param region_name: The AWS region name where the SQS queue
                        containing the S3 event notifications is
                        configured.
    :type region_name: string
    :param queue_name: The name of the SQS queue containing the S3
                       event notifications.
    :type queue_name: string
    """

    def __init__(self, region_name, queue_name):
        self._sqs_handler = SqsHandler(region_name, queue_name)
        self._s3_handler = S3Handler()
        self._results_uploader = []

    def register_results_uploader(self, results_uploader):
        """Registers results uploader.

        Results uploader will upload the analysis results and the
        summary to a specific destination after the analysis is
        finished.
        """
        self._results_uploader.append(results_uploader)

    def run(self):
        """Fetches the OSXCollector output from an S3 bucket based on
        the S3 ObjectCreated event notifications and runs the Analyze
        Filter on the output file.
        Once the analysis is finished the output and the "very readable
        output" files are uploaded to the target S3 bucket.
        """
        created_objects = self._sqs_handler.get_created_objects()

        for created_object in created_objects:
            if (created_object.key_name.endswith('.tar.gz')):
                self._process_created_object(created_object)
            else:
                logging.warn(
                    'S3 object {0} name should end with ".tar.gz"'
                    .format(created_object.key_name))

    def _process_created_object(self, created_object):
        """Fetches the object from an S3 bucket and runs the analysis.
        Then it sends the results to the target S3 bucket and attaches
        them to the JIRA ticket.
        """
        # fetch the OSXCollector output from the S3 bucket
        self._osxcollector_output = self._s3_handler.get_contents_as_string(
            created_object.bucket_name, created_object.key_name)
        self._extract_osxcollector_output_json_file()

        try:
            self._run_analyze_filter()
            self._upload_analysis_results(created_object.key_name)
        except:
            # Log the exception and do not try any recovery.
            # The message that caused the exception will be deleted from the
            # SQS queue to prevent the same exception from happening in the
            # future.
            logging.exception(
                'Unexpected error while running the Analyze Filter')

    def _extract_osxcollector_output_json_file(self):
        """Extracts JSON file containing the OSXCollector output from
        tar.gz archive. It will look in the archive contents for the
        file with the extension ".json". If no file with this extension
        is found in the archive or more than one JSON file is found, it
        will raise `OSXCollectorOutputExtractionError`.
        """
        # create a file-like object based on the S3 object contents as string
        fileobj = StringIO(self._osxcollector_output)
        tar = tarfile.open(mode='r:gz', fileobj=fileobj)
        json_tarinfo = [t for t in tar if t.name.endswith('.json')]

        if 1 != len(json_tarinfo):
            raise OSXCollectorOutputExtractionError(
                'Expected 1 JSON file inside the OSXCollector output archive, '
                'but found {0} instead.'.format(len(json_tarinfo)))

        tarinfo = json_tarinfo[0]
        self._osxcollector_output_json_file = tar.extractfile(tarinfo)
        logging.info(
            'Extracted OSXCollector output JSON file {0}'.format(tarinfo.name))

    def _run_analyze_filter(self):
        """Runs Analyze Filter on the OSXCollector output retrieved
        from an S3 bucket.
        """
        self._analysis_output = StringIO()
        self._text_analysis_summary = StringIO()
        self._html_analysis_summary = StringIO()

        analyze_filter = AnalyzeFilter(
            monochrome=True,
            text_output_file=self._text_analysis_summary,
            html_output_file=self._html_analysis_summary)

        _run_filter(
            analyze_filter,
            input_stream=self._osxcollector_output_json_file,
            output_stream=self._analysis_output)

        # rewind the output files
        self._analysis_output.seek(0)
        self._text_analysis_summary.seek(0)
        self._html_analysis_summary.seek(0)

    def _upload_analysis_results(self, osxcollector_output_filename):
        # drop the file extension (".tar.gz")
        filename_without_extension = osxcollector_output_filename[:-7]

        analysis_output_filename = '{0}_analysis.json'.format(
            filename_without_extension)
        text_analysis_summary_filename = '{0}_summary.txt'.format(
            filename_without_extension)
        html_analysis_summary_filename = '{0}_summary.html'.format(
            filename_without_extension)

        results = [
            FileMetaInfo(
                osxcollector_output_filename,
                StringIO(self._osxcollector_output), 'application/gzip'),
            FileMetaInfo(
                analysis_output_filename, self._analysis_output,
                'application/json'),
            FileMetaInfo(
                text_analysis_summary_filename, self._text_analysis_summary,
                'text'),
            FileMetaInfo(
                html_analysis_summary_filename, self._html_analysis_summary,
                'text/html; charset=UTF-8'),
        ]

        for results_uploader in self._results_uploader:
            results_uploader.upload_results(results)


class OSXCollectorOutputExtractionError(Exception):
    """Raised when an unexpected number of JSON files is found in the
    OSXCollector output archive.
    """

    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import base64
import logging

import requests
from jira.client import JIRA
from requests.exceptions import RequestException

from amira.results_uploader import ResultsUploader


class JiraResultsUploader(ResultsUploader):
    """Attaches the analysis results to a JIRA issue.

    :param jira_server: The JIRA server URL.
    :type jira_server: string
    :param jira_user: The JIRA user.
    :type jira_user: string
    :param jira_password: The JIRA password.
    :type jira_password: string
    :param jira_project: The JIRA project which will be used when
                         looking for the issue related to the
                         OSXCollector output.
    :type jira_project: string
    :param jira_initial_status: The JIRA status which will be used when
                                looking for the issue related to the
                                OSXCollector output.
    :type jira_initial_status: string
    :param jira_final_status: The JIRA status to which the issues will
                              be moved after the analysis.
    :type jira_final_status: string
    """

    def __init__(
            self, jira_server, jira_user, jira_password, jira_project,
            jira_initial_status, jira_final_status):
        self._jira = JIRA(
            jira_server, basic_auth=(jira_user, jira_password))

        self._server = jira_server
        self._project = jira_project
        self._initial_status = jira_initial_status
        self._final_status = jira_final_status

        # create the headers used when uploading the issue attachment
        authorization = base64.b64encode(
            '{0}:{1}'.format(jira_user, jira_password))
        self._headers = {
            'Authorization': 'Basic {0}'.format(authorization),
            'X-Atlassian-Token': 'nocheck'
        }

    def upload_results(self, results):
        """Attaches the analysis results to a JIRA issue determined by
        the file names of the analysis results,
        e.g. for ``name`` == "jdoe-2016_04_05-17_27_38_summary.txt",
        the JIRA issue will be found by looking for the issue in the
        ``jira_project`` project with the ``jira_initial_status``
        status which has the "Affected User" field set to "jdoe".

        :param results: The list containing the meta info (name,
                        content and content-type) of the files which
                        needs to be uploaded.
        :type results: list of ``FileMetaInfo`` tuples

        :raises: JiraIssueNotFoundError
        :raises: JiraFinalStatusNotFoundError
        """
        for file_meta_info in results:
            logging.info(
                'Attaching the file "{0}" to the JIRA issue.'
                .format(file_meta_info.name))
            self._add_attachment(
                file_meta_info.name,
                file_meta_info.content,
                file_meta_info.content_type)

        # as the for loop above will always iterate over few elements,
        # 'jira_issue_key' should be always initialized
        self._move_to_final_status()

    def _add_attachment(self, name, content, content_type):
        """Attaches the file-like object ``content`` to the JIRA issue
        under the name ``name``. ``name`` is also used to look up the
        JIRA issue to which the file will be attached.
        E.g. for ``name`` = "jdoe-2016_04_05-17_27_38_summary.txt",

        :param name: The attachment name.
        :type name: string
        :param content: The attachment as a file-like object.
        :type content: file
        :param content_type: MIME type of the attachment file,
                             e.g. ``text/html``
        :type content_type: string

        :raises: JiraIssueNotFoundError
        """
        # extract the affected user name from the attachment name
        affected_user = name.split('-', 1)[0]
        self._find_issue_by_affected_user(affected_user)

        # do not use jira Python module as it has some problems
        # with handling larger files
        url = '{0}/rest/api/2/issue/{1}/attachments'.format(
            self._server, self._jira_issue_key)
        files = {'file': (name, content.getvalue(), content_type)}

        try:
            requests.post(url, headers=self._headers, files=files)
        except RequestException:
            logging.exception(
                'Error occurred when adding the attachment file {0}'
                ' to the JIRA issue {1}'
                .format(name, self._jira_issue_key))

    def _find_issue_by_affected_user(self, affected_user):
        """Looks up the JIRA issue which "Affected User" field value
        equals `affected_user` parameter value. Only the issues
        belonging to the `jira_project` project and with the
        `jira_status` status will be considered when looking up the
        issue.

        :raises: JiraIssueNotFoundError
        """
        issues = self._jira.search_issues(
            'project={0} AND status={1} AND "Affected User"={2}'
            .format(self._project, self._initial_status, affected_user))

        if not issues:
            raise JiraIssueNotFoundError(
                'No JIRA issue found for the affected user "{0}".'.format(
                    affected_user))

        # return the first JIRA issue found, ignoring the others if any
        self._jira_issue_key = issues[0].key

    def _move_to_final_status(self):
        """Moves the issue to the final status.

        :raises: JiraFinalStatusNotFoundError
        """
        transition = self._find_transition()
        self._jira.transition_issue(self._jira_issue_key, transition)

    def _find_transition(self):
        """Finds the transition to the final status.

        :raises: JiraFinalStatusNotFoundError
        """
        transitions = self._jira.transitions(self._jira_issue_key)
        for transition in transitions:
            if self._final_status == transition['name']:
                return transition['id']

        # no transition found, raise an exception
        raise JiraFinalStatusNotFoundError(
            'Transition to final status "{0}" not found'.format(
                self._final_status))


class JiraIssueNotFoundError(Exception):
    """Raised if JIRA issue cannot be found."""

    pass


class JiraFinalStatusNotFoundError(Exception):
    """Raised if the final status to which the issue was supposed to be
    transitioned cannot be found.
    """

    pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import logging
from collections import namedtuple

import boto.sqs
import simplejson
from boto.sqs.message import RawMessage


# 10 is the maximum number of messages to read at once:
# http://docs.aws.amazon.com/AWSSimpleQueueService/latest/APIReference/API_ReceiveMessage.html
MAX_NUMBER_MESSAGES = 10


CreatedObject = namedtuple('ObjectCreated', ['bucket_name', 'key_name'])


class SqsHandler():
    """Retrieves the S3 event notifications about the objects created
    in the bucket for which the notifications were configured.

    :param region_name: The AWS region name where the SQS queue
                        containing the S3 event notifications is
                        configured.
    :type region_name: string
    :param queue_name: The name of the SQS queue containing the S3
                       event notifications.
    :type queue_name: string
    """

    def __init__(self, region_name, queue_name):
        self._setup_sqs_queue(region_name, queue_name)

    def _setup_sqs_queue(self, region_name, queue_name):
        """Connects to the SQS queue in a given AWS region.

        :param region_name: The AWS region name.
        :type region_name: string
        :param queue_name: The SQS queue name.
        :type queue_name: string
        """
        sqs_connection = boto.sqs.connect_to_region(region_name)
        self.sqs_queue = sqs_connection.get_queue(queue_name)

        if not self.sqs_queue:
            raise SqsQueueNotFoundException(queue_name)

        logging.info('Successfully connected to {0} SQS queue'.format(
            queue_name))

        self.sqs_queue.set_message_class(RawMessage)

    def get_created_objects(self):
        """Retrieves the S3 event notifications about the objects
        created in the OSXCollector output bucket yields the (bucket
        name, key name) pairs describing these objects.
        """
        messages = self.sqs_queue.get_messages(MAX_NUMBER_MESSAGES)
        logging.info('Received {0} message(s) from the SQS queue'.format(
            len(messages)))

        if messages:
            for message in messages:
                objects_created = self._retrieve_created_objects_from_message(
                    message)

                for object_created in objects_created:
                    yield object_created

            self.sqs_queue.delete_message_batch(messages)

    def _retrieve_created_objects_from_message(self, message):
        """Retrieves the bucket name and the key name, describing the
        created object, from the `Records` array in the SQS message.

        Yields each (bucket name, key name) pair as an `CreatedObject`
        named tuple.

        :param message: The SQS message. It should be in the JSON
                        format.
        :type message: string
        """
        json_body = message.get_body()
        body = simplejson.loads(json_body)

        if 'Records' not in body:
            logging.warn(
                '"Records" field not found in the SQS message. '
                'Message body: {0}'.format(body))
            return []

        records = body['Records']
        return self._extract_created_objects_from_records(records)

    def _extract_created_objects_from_records(self, records):
        logging.info(
            'Found {0} record(s) in the SQS message'.format(len(records)))

        for record in records:
            bucket_name = record['s3']['bucket']['name']
            key_name = record['s3']['object']['key']
            yield CreatedObject(bucket_name=bucket_name, key_name=key_name)


class SqsQueueNotFoundException(Exception):
    """An exception thrown when the SQS queue cannot be found."""

    def __init__(self, queue_name):
        self.queue_name = queue_name

    def __str__(self):
        return "SQS queue {0} not found.".format(self.queue_name)
